#include "convert.h"
#include "data.h"
#include "toc.h"
#include "main.h"

#pragma warning(disable:4996)

HWND g_hConvertWnd = NULL;

int getsize(FILE *f)
{
	int size;

	fseek(f, 0, SEEK_END);
	size = ftell(f);

	fseek(f, 0, SEEK_SET);
	return size;
}

void ErrorExit(char *fmt, ...)
{
	va_list list;
	char msg[256];	

	va_start(list, fmt);
	vsprintf(msg, fmt, list);
	va_end(list);

	printf(msg);
	exit(-1);
}

z_stream z;

int deflateCompress(void *inbuf, int insize, void *outbuf, int outsize, int level)
{
	int res;
	
	z.zalloc = Z_NULL;
	z.zfree  = Z_NULL;
	z.opaque = Z_NULL;

	if (deflateInit2(&z, level , Z_DEFLATED, -15, 8, Z_DEFAULT_STRATEGY) != Z_OK)
		return -1;

	z.next_out  = (Bytef *)outbuf;
	z.avail_out = outsize;
	z.next_in   = (Bytef *)inbuf;
	z.avail_in  = insize;

	if (deflate(&z, Z_FINISH) != Z_STREAM_END)
	{
		return -1;
	}

	res = outsize - z.avail_out;

	if (deflateEnd(&z) != Z_OK)
		return -1;

	return res;
}

#ifndef WIN32
typedef struct __attribute__((packed))
#else
#pragma pack(1)
typedef struct
#endif
{
	unsigned int signature;
	unsigned int version;
	unsigned int fields_table_offs;
	unsigned int values_table_offs;
	int nitems;
} SFOHeader;

#ifndef WIN32
typedef struct __attribute__((packed))
#else
#pragma pack(1)
typedef struct
#endif
{
	unsigned short field_offs;
	unsigned char  unk;
	unsigned char  type; // 0x2 -> string, 0x4 -> number
	unsigned int length;
	unsigned int size;
	unsigned short val_offs;
	unsigned short unk4;
} SFODir;

//SFOf[^---------------------------------------------------------------
char sfo_header[8]={0x00,0x50,0x53,0x46,0x01,0x01,0x00,0x00};//\0PSF\1\1\0\0

typedef enum SFO_INDEX
{
	BOOTABLE,
	CATEGORY,
	DISC_ID,
	DISC_VERSION,
	LICENSE,
	PARENTAL_LEVEL,
	PSP_SYSTEM_VER,
	REGION,
	TITLE
};

const int pm_n = 9;
char pm_title[pm_n][0x10]={
	"BOOTABLE",
	"CATEGORY",
	"DISC_ID",
	"DISC_VERSION",
	"LICENSE",
	"PARENTAL_LEVEL",
	"PSP_SYSTEM_VER",
	"REGION",
	"TITLE"
};
char pm_type[pm_n]={1,0,0,0,0,1,0,1,0};
int pm_len[pm_n]={0x4,0x4,0x10,0x8,0x200,0x4,0x8,0x4,0x80};
int pm_b[pm_n]={1,0,0,0,0,0,0,0,0};
char pm_s[pm_n][0x200]={
	"",
	"ME",
	"",
	"1.00",
	"PopstationMD Free GUI",
	"",
	"3.71",
	"",
	""
};
//-----------------------------------------------------------------------
int CreateSFO(char* szFilePath){
	FILE *out = fopen(szFilePath,"wb");
	if(!out){
		fclose(out);
		return 0;		
	}
	
	//PARAM.SFO쐬
    int sa=0;
    int po=0;
	int i=0;
    fwrite(sfo_header,1,sizeof(sfo_header),out);
        
    int str_ad=sizeof(sfo_header)+4+4+4+16*pm_n;
    fwrite(&str_ad,1,4,out);
    int byte_ad=0;
    for(i=0;i<pm_n;i++){
    	byte_ad+=(int)strlen(pm_title[i])+1;
    }
        
    while((str_ad+byte_ad)%0x4!=0){
        byte_ad++;
    	sa++;
    }
    po=(str_ad+byte_ad);
    fwrite(&po,1,4,out);
    fwrite(&pm_n,1,4,out);
    
    short str_offset=0;
    int byte_offset=0;
    
    for(i=0;i<pm_n;i++){
        fwrite(&str_offset,1,2,out);
        po=4;
        fwrite(&po,1,1,out);
        if(pm_type[i]){
        	po=4;
        	fwrite(&po,1,1,out);
        }else{
        	po=2;
        	fwrite(&po,1,1,out);
        }
        fwrite(&pm_len[i],1,4,out);
        fwrite(&pm_len[i],1,4,out);
        fwrite(&byte_offset,1,4,out);
        
        str_offset+=(short)strlen(pm_title[i])+1;
        byte_offset+=pm_len[i];
    }
    for(i=0;i<pm_n;i++){
    	fwrite(pm_title[i],1,strlen(pm_title[i])+1,out);
    }
    for(i=0;i<sa;i++){
    	fputc(0, out);
    }
    for(i=0;i<pm_n;i++){
        if(pm_type[i]){
        	fwrite(&pm_b[i],1,pm_len[i],out);
        }else{
        	fwrite(pm_s[i],1,pm_len[i],out);
        }
    }
	
	fclose(out);
	return 1;
}

char buffer[1*1048576];
char buffer2[0x9300];

int sfo = 0,pic0 = 0, pic1 = 0, icon0 = 0, icon1 = 0, snd = 0, prx = 0, toc = 0, splash = 0;
int sfo_size, pic0_size, pic1_size, icon0_size, icon1_size, snd_size, prx_size, toc_size, splash_size;
int start_dat = 0;
unsigned int header[0x28/4];
unsigned int dummy[6];

typedef struct IsoIndex
{
	unsigned int offset;
	unsigned int length;
	unsigned int dummy[6];
} IsoIndex;

#define WriteInteger(a, c) \
	x = a; \
	for (i = 0; i < c; i++) \
		fwrite(&x, 1, 4, out)

#define WriteChar(a, c) \
	for (i = 0; i < c; i++) \
		fputc(a, out)

#define WriteRandom(c) \
	for (i = 0; i < c; i++) \
	{ \
		x = rand(); \
		fwrite(&x, 1, 4, out); \
	}

PDATA g_pConvertData = NULL;
HANDLE g_hThread = NULL;
DWORD g_dwThreadId = NULL; 

int getTotalPos(int progDiscPos[MAX_DISC_NUM]) {
	int totalPos = 0;

	for (int i = 0; i < MAX_DISC_NUM; i++) {
		totalPos += progDiscPos[i];
	}
	
	return totalPos;
}

bool convert(PDATA pData, char *output, char *title, char titles[MAX_DISC_NUM][0x80], char *code, int *complevels)
{
	char icon0_path[MAX_PATH], icon1_path[MAX_PATH], pic0_path[MAX_PATH], pic1_path[MAX_PATH], snd_path[MAX_PATH], splash_path[MAX_PATH];
	char prx_path[MAX_PATH], sfo_path[MAX_PATH];

	FILE *in, *out, *t;
	int i, x, isosize, isorealsize;
	__int64 offset, end_x;
	__int64 index_offset, p1_offset, p2_offset, m_offset, end_offset;
	IsoIndex *indexes;
	unsigned int iso_positions[5];
	int ciso;
	int read, written;
	int progDiscPos[MAX_DISC_NUM];
	char message[1024];

	void* tocptr;

	bool bResult = true;

	// t@CpX̏
	memset(icon0_path, 0, MAX_PATH);
	memset(icon1_path, 0, MAX_PATH);
	memset(pic0_path, 0, MAX_PATH);
	memset(pic1_path, 0, MAX_PATH);
	memset(snd_path, 0, MAX_PATH);
	memset(splash_path, 0, MAX_PATH);

	for (i = 0; i < MAX_DISC_NUM; i++) {
		progDiscPos[i] = 0;
	}

	SendDlgItemMessage(pData->hwnd, IDC_PROGRESS_TOTAL, PBM_SETRANGE, 0, (LPARAM)MAKELONG(0,100 * pData->ndiscs));
	SendDlgItemMessage(pData->hwnd, IDC_PROGRESS_DISC, PBM_SETRANGE, 0, (LPARAM)MAKELONG(0, 100));
	SendDlgItemMessage(pData->hwnd, IDC_PROGRESS_TOTAL, PBM_SETPOS, 0, 0);
	SendDlgItemMessage(pData->hwnd, IDC_PROGRESS_DISC, PBM_SETPOS, 0, 0);

	// ACRi[tH_̃t@Cǂݍ
	if (strlen(pData->icon_dir) > 0) {
		strcpy(icon0_path, pData->icon_dir);
		strcat(icon0_path, "ICON0.PNG");
		t = fopen(icon0_path, "rb");
		if (t)
		{
			icon0_size = getsize(t);
			icon0 = 1;
			fclose(t);
		}
		else
		{
			strcpy(icon0_path, g_exec_dir);
			strcat(icon0_path, "ICON0.PNG");
		}

		strcpy(icon1_path, pData->icon_dir);
		strcat(icon1_path, "ICON1.PMF");
		t = fopen(icon1_path, "rb");
		if (t)
		{
			icon1_size = getsize(t);
			icon1 = 1;
			fclose(t);
		}
		else
		{
			strcpy(icon1_path, g_exec_dir);
			strcat(icon1_path, "ICON1.PMF");
		}

		strcpy(pic0_path, pData->icon_dir);
		strcat(pic0_path, "PIC0.PNG");
		t = fopen(pic0_path, "rb");
		if (t)
		{
			pic0_size = getsize(t);
			pic0 = 1;
			fclose(t);
		}
		else
		{
			strcpy(pic0_path, g_exec_dir);
			strcat(pic0_path, "PIC0.PNG");
		}

		strcpy(pic1_path, pData->icon_dir);
		strcat(pic1_path, "PIC1.PNG");
		t = fopen(pic1_path, "rb");
		if (t)
		{
			pic1_size = getsize(t);
			pic1 = 1;
			fclose(t);
		}
		else
		{
			strcpy(pic1_path, g_exec_dir);
			strcat(pic1_path, "PIC1.PNG");
		}

		strcpy(snd_path, pData->icon_dir);
		strcat(snd_path, "SND0.AT3");
		t = fopen(snd_path, "rb");
		if (t)
		{
			snd_size = getsize(t);
			snd = 1;
			fclose(t);
		}
		else
		{
			strcpy(snd_path, g_exec_dir);
			strcat(snd_path, "SND0.AT3");
		}

		strcpy(splash_path, pData->icon_dir);
		strcat(splash_path, "SPLASH.PNG");
		t = fopen(splash_path, "rb");
		if (t){
			splash_size = getsize(t);
			splash = 1;
			fclose(t);
		}else{
			strcpy(splash_path, g_exec_dir);
			strcat(splash_path, "SPLASH.PNG");
		}
	}
	else
	{
		strcpy(icon0_path, g_exec_dir);
		strcat(icon0_path, "ICON0.PNG");
		strcpy(icon1_path, g_exec_dir);
		strcat(icon1_path, "ICON1.PMF");
		strcpy(pic0_path, g_exec_dir);
		strcat(pic0_path, "PIC0.PNG");
		strcpy(pic1_path, g_exec_dir);
		strcat(pic1_path, "PIC1.PNG");
		strcpy(snd_path, g_exec_dir);
		strcat(snd_path, "SND0.AT3");
		strcpy(splash_path, g_exec_dir);
		strcat(splash_path, "SPLASH.PNG");
	}

	if (icon0 == 0)
	{
		t = fopen(icon0_path, "rb");
		if (t)
		{
			icon0_size = getsize(t);
			icon0 = 1;
			fclose(t);
		}
		else
		{
			icon0_size = 0;

			writeLog(pData->hwnd, "ICON0.PNG܂ł");
			return false;
		}
	}

	if (icon1 == 0)
	{
		t = fopen(icon1_path, "rb");
		if (t)
		{
			icon1_size = getsize(t);
			icon1 = 1;
			fclose(t);
		}
		else
		{
			icon1_size = 0;
		}
	}

	if (pic0 == 0)
	{
		t = fopen(pic0_path, "rb");
		if (t)
		{
			pic0_size = getsize(t);
			pic0 = 1;
			fclose(t);
		}
		else
		{
			pic0_size = 0;
		}
	}

	if (pic1 == 0)
	{
		t = fopen(pic1_path, "rb");
		if (t)
		{
			pic1_size = getsize(t);
			pic1 = 1;
			fclose(t);
		}
		else
		{
			pic1_size = 0;
		}
	}

	if (snd == 0)
	{
		t = fopen(snd_path, "rb");
		if (t)
		{
			snd_size = getsize(t);
			snd = 1;
			fclose(t);
		}
		else
		{
			snd = 0;
		}
	}

	strncpy(prx_path, g_exec_dir, MAX_PATH);
	strncat(prx_path, "DATA.PSP", MAX_PATH); 
	t = fopen(prx_path, "rb");
	if (t)
	{
		prx_size = getsize(t);
		prx = 1;
		fclose(t);
	}
	else
	{
		prx_size = 0;

		writeLog(pData->hwnd, "DATA.PSP܂ł");
		return false;
	}

	if (splash == 0)
	{
		t = fopen(splash_path, "rb");
		if (t){
			splash_size = getsize(t);
			splash = 1;
			fclose(t);
		}else{
			splash_size = 0;

			writeLog(pData->hwnd, "SPLASH.PNG܂ł");
			return false;
		}
	}

	strncpy(sfo_path, g_exec_dir, MAX_PATH);
	strncat(sfo_path, "PARAM.SFO", MAX_PATH); 
	if (CreateSFO(sfo_path) == 1){
		t = fopen(sfo_path, "rb");
		sfo_size = getsize(t);
		sfo = 1;
		fclose(t);
	}

	out = fopen(output, "wb");
	if (!out)
	{
		sprintf(message, "%s쐬邱Ƃł܂\n", output);
		writeLog(pData->hwnd, message);
		return false;
	}

	int curoffs = 0x28;

	header[0] = 0x50425000;
	header[1] = 0x10000;
	
	header[2] = curoffs;

	curoffs += sfo_size;
	header[3] = curoffs;

	curoffs += icon0_size;
	header[4] = curoffs;

	curoffs += icon1_size;
	header[5] = curoffs;

	curoffs += pic0_size;
	header[6] = curoffs;

	curoffs += pic1_size;
	header[7] = curoffs;

	curoffs += snd_size;
	header[8] = curoffs;

	x = header[8] + prx_size;

	if ((x % 0x10000) != 0)
	{
		x = x + (0x10000 - (x % 0x10000));
	}
	
	header[9] = x;

	fwrite(header, 1, 0x28, out);

	writeLog(pData->hwnd, "SFOł܂...", FALSE);

	if (sfo)
	{
		t = fopen(sfo_path, "rb");
    	fread(buffer, 1, sfo_size, t);
    	fwrite(buffer, 1, sfo_size, out);
    	fclose(t);
		remove(sfo_path);

		writeLog(pData->hwnd, "");
	} else {
		writeLog(pData->hwnd, "G[");
		fclose(out);
		return false;
	}

	if (icon0)
	{
		writeLog(pData->hwnd, "ICON0.PNGł܂...", FALSE);

		t = fopen(icon0_path, "rb");
		fread(buffer, 1, icon0_size, t);
		fwrite(buffer, 1, icon0_size, out);
		fclose(t);

		writeLog(pData->hwnd, "");
	}

	if (icon1)
	{
		writeLog(pData->hwnd, "ICON1.PMFł܂...", FALSE);
		
		t = fopen(icon1_path, "rb");
		fread(buffer, 1, icon1_size, t);
		fwrite(buffer, 1, icon1_size, out);
		fclose(t);

		writeLog(pData->hwnd, "");
	}

	if (pic0)
	{
		writeLog(pData->hwnd, "PIC0.PNGł܂...", FALSE);
		
		t = fopen(pic0_path, "rb");
		fread(buffer, 1, pic0_size, t);
		fwrite(buffer, 1, pic0_size, out);
		fclose(t);

		writeLog(pData->hwnd, "");
	}

	if (pic1)
	{
		writeLog(pData->hwnd, "PIC1.PNGł܂...", FALSE);
		
		t = fopen(pic1_path, "rb");
		fread(buffer, 1, pic1_size, t);
		fwrite(buffer, 1, pic1_size, out);
		fclose(t);

		writeLog(pData->hwnd, "");
	}

	if (snd)
	{
		writeLog(pData->hwnd, "SND0.AT3ł܂...", FALSE);
		
		t = fopen(snd_path, "rb");
		fread(buffer, 1, snd_size, t);
		fwrite(buffer, 1, snd_size, out);
		fclose(t);

		writeLog(pData->hwnd, "");
	}

	if (prx)
	{
		writeLog(pData->hwnd, "DATA.PSPł܂...", FALSE);

		t = fopen(prx_path, "rb");
		fread(buffer, 1, prx_size, t);
		fwrite(buffer, 1, prx_size, out);
		fclose(t);

		writeLog(pData->hwnd, "");
	}

	offset = _ftelli64(out);
	
	for (i = 0; i < header[9]-offset; i++)
	{
		fputc(0, out);
	}

	if (pData->ndiscs > 1)
	{
		writeLog(pData->hwnd, "PSTITLEwb_ł܂...", FALSE);

		fwrite("PSTITLEIMG000000", 1, 16, out);

		// Save this offset position
		p1_offset = _ftelli64(out);
		
		WriteInteger(0, 2);
		WriteInteger(0x2CC9C5BC, 1);
		WriteInteger(0x33B5A90F, 1);
		WriteInteger(0x06F6B4B3, 1);
		WriteInteger(0xB25945BA, 1);
		WriteInteger(0, 0x76);

		m_offset = _ftelli64(out);

		memset(iso_positions, 0, sizeof(iso_positions));
		fwrite(iso_positions, 1, sizeof(iso_positions), out);

		WriteRandom(12);
		WriteInteger(0, 8);
		
		fputc('_', out);
		fwrite(code, 1, 4, out);
		fputc('_', out);
		fwrite(code+4, 1, 5, out);

		WriteChar(0, 0x15);
		
		p2_offset = _ftelli64(out);
		WriteInteger(0, 2);
		
		fwrite(data3, 1, sizeof(data3), out);
		fwrite(title, 1, strlen(title), out);

		WriteChar(0, 0x80 - (int)strlen(title));
		WriteInteger(7, 1);
		WriteInteger(0, 0x1C);

		writeLog(pData->hwnd, "");
	}

	for (ciso = 0; ciso < pData->ndiscs; ciso++)
	{
		sprintf(message, "DISC%d", ciso + 1);
		SendDlgItemMessage(pData->hwnd, IDC_LABEL_DISC, WM_SETTEXT, 0, (LPARAM)message);
		SendDlgItemMessage(pData->hwnd, IDC_PROGRESS_DISC, PBM_SETPOS, 0, 0);

		in = fopen (pData->inputs[ciso], "rb");
		if (!in)
		{
			sprintf(message, "%sJ܂\n", pData->inputs[ciso]);
			writeLog(pData->hwnd, message);
			fclose(out);
			return false;
		}

		if ((tocptr = create_toc(pData->inputs[ciso], &toc_size, pData->hwnd)) != NULL)
		{
			toc = 2;
		}
		else
		{
			toc = 0;
		}

		isosize = getsize(in);
		isorealsize = isosize;

		if ((isosize % 0x9300) != 0)
		{
			isosize = isosize + (0x9300 - (isosize%0x9300));
		}
		
		offset = _ftelli64(out);

		if (offset % 0x8000)
		{
			x = (int)(0x8000 - (offset % 0x8000));

			WriteChar(0, x);			
		}

		iso_positions[ciso] = (unsigned int)(_ftelli64(out) - header[9]);

		fwrite("PSISOIMG0000", 1, 12, out);

		if (pData->ndiscs > 1)
		{
			WriteInteger(0, 0xFD);
		}
		else
		{
			p1_offset = _ftelli64(out);

			x = isosize + 0x100000;
			fwrite(&x, 1, 4, out);
			WriteInteger(0, 0xFC);
		}

		memcpy(data1+1, pData->discids[ciso], 4);
		memcpy(data1+6, pData->discids[ciso]+4, 5);

		if (toc == 2){
			memcpy(data1+1024, tocptr, toc_size);
			free(tocptr);
		}

		fwrite(data1, 1, sizeof(data1), out);

		if (pData->ndiscs > 1)
		{
			WriteInteger(0, 1);
		}
		else
		{
			p2_offset = _ftelli64(out);
			x = isosize + 0x100000 + 0x2d31;
			fwrite(&x, 1, 4, out);
		}
		
		strcpy((char *)data2+8, titles[ciso]);
		fwrite(data2, 1, sizeof(data2), out);

		index_offset = _ftelli64(out);

		//printf("Writing indexes (iso #%d)...\n", ciso+1);

		memset(dummy, 0, sizeof(dummy));

		offset = 0;

		if (complevels[ciso] == 0)
		{	
			x = 0x9300;
		}
		else
		{
			x = 0;
		}

		for (i = 0; i < isosize / 0x9300; i++)
		{
			fwrite(&offset, 1, 4, out);
			fwrite(&x, 1, 4, out);
			fwrite(dummy, 1, sizeof(dummy), out);

			if (complevels[ciso] == 0)
				offset += 0x9300;
		}

		offset = _ftelli64(out);

		for (i = 0; i < (iso_positions[ciso]+header[9]+0x100000)-offset; i++)
		{
			fputc(0, out);
		}

		sprintf(message, "DISC%d̃C[Wt@Cł܂...", ciso+1);
		writeLog(pData->hwnd, message, FALSE);

		written = 0;

		if (complevels[ciso] == 0)
		{
			while ((x = (int)fread(buffer, 1, 1048576, in)) > 0)
			{
				fwrite(buffer, 1, x, out);
				written += x;

				progDiscPos[ciso] = (int)(((double)written / (double)isorealsize) * (double)100);
				SendDlgItemMessage(pData->hwnd, IDC_PROGRESS_DISC, PBM_SETPOS, progDiscPos[ciso], 0);
				SendDlgItemMessage(pData->hwnd, IDC_PROGRESS_TOTAL, PBM_SETPOS, getTotalPos(progDiscPos), 0);

				if (pData->thread_end == TRUE) {
					fclose(in);
					fclose(out);

					writeLog(pData->hwnd, "~");
					return false;
				}
			}

			for (i = 0; i < (isosize-isorealsize); i++)
			{
				fputc(0, out);
			}

			progDiscPos[ciso] = 100;
			SendDlgItemMessage(pData->hwnd, IDC_PROGRESS_DISC, PBM_SETPOS, progDiscPos[ciso], 0);
			SendDlgItemMessage(pData->hwnd, IDC_PROGRESS_TOTAL, PBM_SETPOS, getTotalPos(progDiscPos), 0);

			writeLog(pData->hwnd, "");
		}
		else
		{
			indexes = (IsoIndex *)malloc(sizeof(IsoIndex) * (isosize/0x9300));

			if (!indexes)
			{
				fclose(in);
				fclose(out);

				writeLog(pData->hwnd, "G[");
				writeLog(pData->hwnd, "CfbNX̂߂̃蓖Ă邱Ƃł܂ł");
				return false;
			}

			i = 0;
			offset = 0;
			read = 0;

			while ((x = (int)fread(buffer2, 1, 0x9300, in)) > 0)
			{
				if (x < 0x9300)
				{
					memset(buffer2+x, 0, 0x9300-x);
				}
			
				read = x;
				x = deflateCompress(buffer2, 0x9300, buffer, sizeof(buffer), complevels[ciso]);

				if (x < 0)
				{
					fclose(in);
					fclose(out);
					free(indexes);

					writeLog(pData->hwnd, "G[");
					writeLog(pData->hwnd, "kŃG[܂");
					return false;
				}

				memset(&indexes[i], 0, sizeof(IsoIndex));

				indexes[i].offset = (unsigned int)offset;

				if (x >= 0x9300) /* Block didn't compress */
				{				
					indexes[i].length = 0x9300;
					fwrite(buffer2, 1, 0x9300, out);
					offset += 0x9300;
				}
				else
				{
					indexes[i].length = x;
					fwrite(buffer, 1, x, out);
					offset += x;
				}

				written += read;
				progDiscPos[ciso] = (int)(((double)written / (double)isosize) * (double)100);
				SendDlgItemMessage(pData->hwnd, IDC_PROGRESS_DISC, PBM_SETPOS, progDiscPos[ciso], 0);
				SendDlgItemMessage(pData->hwnd, IDC_PROGRESS_TOTAL, PBM_SETPOS, getTotalPos(progDiscPos), 0);

				i++;

				if (pData->thread_end == TRUE) {
					fclose(in);
					fclose(out);
					free(indexes);

					writeLog(pData->hwnd, "~");
					return false;
				}
			}

			progDiscPos[ciso] = 100;
			SendDlgItemMessage(pData->hwnd, IDC_PROGRESS_DISC, PBM_SETPOS, progDiscPos[ciso], 0);
			SendDlgItemMessage(pData->hwnd, IDC_PROGRESS_TOTAL, PBM_SETPOS, getTotalPos(progDiscPos), 0);

			writeLog(pData->hwnd, "");

			if (i != (isosize/0x9300))
			{
				fclose(in);
				fclose(out);
				free(indexes);

				writeLog(pData->hwnd, "G[");
				writeLog(pData->hwnd, "炩̃G[܂");
				return false;
			}
		}

		if (pData->ndiscs > 1)
		{
			if (complevels[ciso] != 0)
			{
				offset = _ftelli64(out);
				
				sprintf(message, "DISC%d̈kł܂...", ciso+1);
				writeLog(pData->hwnd, message, FALSE);

				_fseeki64(out, index_offset, SEEK_SET);
				fwrite(indexes, 1, sizeof(IsoIndex) * (isosize/0x9300), out);

				_fseeki64(out, offset, SEEK_SET);

				writeLog(pData->hwnd, "");
			}
		}
	}

	end_x = _ftelli64(out);

	if ((end_x % 0x10) != 0)
	{
		end_offset = end_x + (0x10 - (end_x % 0x10));
			
		for (i = 0; i < (end_offset - end_x); i++)
		{
			fputc('0', out);
		}
	}
	else
	{
		end_offset = end_x;
	}

	end_offset -= header[9];

	writeLog(pData->hwnd, "SPLASH.PNGł܂...", FALSE);

	fwrite("STARTDAT", 1, 8, out);
	char st_header[8]={0x01,0x00,0x00,0x00,0x01,0x00,0x00,0x00};
	
	fwrite(st_header,1,8,out);
	char splash_ad[4]={0x50,0x00,0x00,0x00};
	fwrite(splash_ad,1,4,out);
	
	fwrite(&splash_size,1,4,out);
	for (i = 0; i < 0x50-24; i++){
         fputc(0, out);
    }
	
	if (splash){
        t = fopen(splash_path, "rb");
        fread(buffer, 1, splash_size, t);
        fwrite(buffer, 1, splash_size, out);
        fclose(t);

		writeLog(pData->hwnd, "");
	}

	if (pData->ndiscs > 1)
	{
		_fseeki64(out, p1_offset, SEEK_SET);
		fwrite(&end_offset, 1, 4, out);

		end_offset += 0x2d31;
		_fseeki64(out, p2_offset, SEEK_SET);
		fwrite(&end_offset, 1, 4, out);

		_fseeki64(out, m_offset, SEEK_SET);
		fwrite(iso_positions, 1, sizeof(iso_positions), out);
	}
	else
	{
		if (complevels[0] != 0){
			writeLog(pData->hwnd, "kł܂...", FALSE);

			_fseeki64(out, p1_offset, SEEK_SET);
			fwrite(&end_offset, 1, 4, out);

			end_offset += 0x2d31;
			_fseeki64(out, p2_offset, SEEK_SET);
			fwrite(&end_offset, 1, 4, out);

			_fseeki64(out, index_offset, SEEK_SET);
			fwrite(indexes, 1, sizeof(IsoIndex) * (isosize/0x9300), out);

			writeLog(pData->hwnd, "");
		}
	}

	fclose(in);
	fclose(out);

	return true;
}

void to_utf8( char* pcWork, char* pcBuf )
{

    wchar_t wszBuf[1600];

    // Char  Unicode
    MultiByteToWideChar(
        CP_ACP,    // ANSI code page
        0,    // ̕ϊ@
        (const char *)pcWork,    // ϊ
        -1,    // ϊTCY(-1wŎvZ)
        (LPWSTR)wszBuf,    // ϊobt@
        1025);    // ϊobt@TCY


    // Unicode  UTF8
    WideCharToMultiByte(
        CP_UTF8,
        0,
        (LPCWSTR)wszBuf,
        -1,
        (char *)pcBuf,
        2048 * 2,
        NULL,
        NULL);

    
    return;
}

void ShowConvertWnd(int nMode)
{
	if (g_hConvertWnd != NULL) {
		if (nMode == 0) {
			ShowWindow(g_hConvertWnd, SW_HIDE);
		} else {
			ShowWindow(g_hConvertWnd, SW_RESTORE);
		}
	}
}

//-------------------------------------------------------------------------------------------------
//      ϊpf[^ݒ
//-------------------------------------------------------------------------------------------------
void SetConvertData(char* title, int ndiscs, char inputs[MAX_DISC_NUM][MAX_PATH],
					char discids[MAX_DISC_NUM][10], char levels[MAX_DISC_NUM][2], char* icon_dir, char* parental, char* region) {
	g_pConvertData = new DATA;

	// ^Cg
	memset(g_pConvertData->title, 0, 0x80);
	strcpy(g_pConvertData->title, title);
	
	// fBXN
	g_pConvertData->ndiscs = ndiscs;

	for (int i = 0; i < ndiscs; i++) {
		// ISOt@C
		strncpy(g_pConvertData->inputs[i], inputs[i], MAX_PATH);

		// fBXNID
		strncpy(g_pConvertData->discids[i], discids[i], 10);

		// kx
		strncpy(g_pConvertData->levels[i], levels[i], 2);
	}

	// R[h
	strncpy(g_pConvertData->parental, parental, 3);

	// [WR[h
	strncpy(g_pConvertData->region, region, 5);

	// ACRi[tH_
	strncpy(g_pConvertData->icon_dir, icon_dir, MAX_PATH);
}

void CloseConvert()
{
	if (g_hConvertWnd)
	{
		SendMessage(g_hConvertWnd, WM_CLOSE, 0, 0);
	}
}

//-------------------------------------------------------------------------------------------------
//      R[obN֐
//-------------------------------------------------------------------------------------------------
LRESULT CALLBACK DlgProcConvert(HWND hWnd, UINT uiMsg, WPARAM wParam, LPARAM lParam)
{
	switch(uiMsg) {
	case WM_CLOSE:
		g_hConvertWnd = NULL;
		delete g_pConvertData;
		EndDialog(hWnd, IDCANCEL);
        break;
	case WM_INITDIALOG:
		OnInitConvert(hWnd);
		return TRUE;
	case WM_CONVERT_START:
		OnConvertStart(hWnd);
		break;
	case WM_COMMAND:
		switch(LOWORD(wParam)) {
		case IDOK:
			return TRUE;
		case IDCANCEL:
			return TRUE;
		case IDC_CANCEL_OR_CLOSE:
			if (g_pConvertData->thread_end == TRUE) {
				PostMessage(hWnd, WM_CLOSE, 0, 0);
			} else {
				OnCancel(hWnd);
			}
			break;
		}
		break;
	default:
		break;
	}

	return FALSE;
}

void OnCancel(HWND hWnd)
{
	int nRet = MessageBox(hWnd, "ϊ𒆎~܂B\r\n̂܂ܕϊʂ܂H", "~mF", MB_ICONQUESTION | MB_YESNOCANCEL);

	if (nRet == IDYES) {
		g_pConvertData->thread_end = TRUE;
		g_pConvertData->close_flg = TRUE;
	} else if (nRet == IDNO) {
		g_pConvertData->thread_end = TRUE;
		g_pConvertData->close_flg = FALSE;
	}
}

void OnInitConvert(HWND hWnd)
{
	// ACR̐ݒ
	HICON hIconSm;
	hIconSm = LoadIcon(g_hInst, MAKEINTRESOURCE(IDI_ICON));
	SendMessage(hWnd, WM_SETICON, ICON_SMALL, (LPARAM)hIconSm);

	g_hConvertWnd = hWnd;

	// EBhEʒuݒ肷
	SetWindowPos(hWnd, HWND_TOP , g_main_wnd_left, g_main_wnd_bottom, 0, 0, SWP_NOSIZE);

	PostMessage(hWnd, WM_CONVERT_START, 0, 0);
}

void OnConvertStart(HWND hWnd) {
 
	g_pConvertData->hwnd = g_hConvertWnd;
	g_pConvertData->thread_end = FALSE;
	g_pConvertData->close_flg = FALSE;

	g_hThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)ConvertThread, (LPVOID)g_pConvertData, 0, &g_dwThreadId);
}

void initParam() {
	pm_b[0] = 1;
	pm_b[1] = 0;
	pm_b[2] = 0;
	pm_b[3] = 0;
	pm_b[4] = 0;
	pm_b[5] = 0;
	pm_b[6] = 0;
	pm_b[7] = 0;
	pm_b[8] = 0;
	
	for (int i = 0; i < pm_n; i++) {
		memset(pm_s[i], 0, pm_len[i]);
	}
	strcpy(pm_s[0], "");
	strcpy(pm_s[1], "ME");
	strcpy(pm_s[2], "");
	strcpy(pm_s[3], "1.00");
	strcpy(pm_s[4], "PopstationMD Free GUI");
	strcpy(pm_s[5], "");
	strcpy(pm_s[6], "3.71");
	strcpy(pm_s[7], "");
	strcpy(pm_s[8], "");

	sfo = 0;
	pic0 = 0;
	pic1 = 0;
	icon0 = 0;
	icon1 = 0;
	snd = 0;
	prx = 0;
	toc = 0;
	splash = 0;
	
	start_dat = 0;
}

DWORD WINAPI ConvertThread(LPVOID data) {
	PDATA pData;
	FILE *info;
	int i, k;
	int complevels[MAX_DISC_NUM];
	char utf_title[0x80];
	char utf_titles[MAX_DISC_NUM][0x80];
	char discid[10];
	char output_dir[MAX_PATH], output_file[MAX_PATH], output_info[MAX_PATH];
	char caption[10];
	BOOL result = FALSE;

	pData = (PDATA)data;

	initParam();

	to_utf8(pData->title, utf_title);
	for (i = 0; i < pData->ndiscs; i++)
	{
		to_utf8(pData->title, utf_titles[i]);
		complevels[i] = pData->levels[i][0] - '0';
	}
	strncpy(discid, pData->discids[0], 10);

	// ^Cg
	for(i = 0; i <= (int)strlen(utf_title); i++){
		pm_s[TITLE][i] = utf_title[i];
	}

	// Q[R[h
	for(i = 0; i < 9; i++){
		pm_s[DISC_ID][i] = pData->discids[0][i];
	}
	pm_s[DISC_ID][9] = 0;

	// R[h
	i = (int)strlen(pData->parental);
	for(k = 0; k < i; k++){
		int b = 1;
		int j = 0;
		for(j = 0; j < i - k - 1; j++) {
			b *= 10;
		}
		pm_b[PARENTAL_LEVEL] += (pData->parental[k] - '0') * b;
	}

	// [WR[h
	i = (int)strlen(pData->region);
	for(k = 0; k < i; k++){
		int b = 1;
		int j = 0;
		for(j = 0; j < i - k - 1; j++) {
			b *= 16;
		}
		pm_b[REGION] += (pData->region[k] - '0') * b;
	}

	// Ώۃo[Wݒ
	if (pData->ndiscs == 1)
	{
		strcpy(pm_s[PSP_SYSTEM_VER], "3.02");
	}

	strcpy(output_dir, g_output_dir);
	strcat(output_dir, discid);
	strcpy(output_file, output_dir);
	strcat(output_file, "\\EBOOT.PBP");
	strcpy(output_info, output_dir);
	strcat(output_info, "\\info.txt");
	result = CreateDirectoryA(output_dir, NULL);
	if (result == TRUE || GetLastError() == ERROR_ALREADY_EXISTS)
	{
		writeLog(pData->hwnd, "ϊJn\r\n");

		if (convert(pData, output_file, utf_title, utf_titles, discid, complevels) == true) {
			info = fopen(output_info, "w");
			if (info)
			{
				fprintf(info, pData->title);
				fprintf(info, "\r\nConvert By ");
				fprintf(info, APP_NAME);
				fclose(info);
			}

			writeLog(pData->hwnd, "\r\nϊI");
			strcpy(caption, "ϊI");
			SendMessage(pData->hwnd, WM_SETTEXT, 0, (LPARAM)caption);
		} else {
			if (pData->thread_end == TRUE) {
				writeLog(pData->hwnd, "\r\nϊ~܂");
				strcpy(caption, "ϊ~");
			} else {
				writeLog(pData->hwnd, "\r\nϊs");
				strcpy(caption, "ϊs");
			}
			SendMessage(pData->hwnd, WM_SETTEXT, 0, (LPARAM)caption);
		}
	}

	strcpy(caption, "");
	SendDlgItemMessage(pData->hwnd, IDC_CANCEL_OR_CLOSE, WM_SETTEXT, 0, (LPARAM)caption);

	pData->thread_end = TRUE;

	if (pData->close_flg == TRUE) {
		PostMessage(pData->hwnd, WM_CLOSE, 0, 0);
	}

	PostMessage(g_hMainWnd, WM_CONVERT_END, 0, 0);

	return (DWORD)0; 
}

void writeLog(HWND hWnd, char* message, BOOL bNextLine) {
	LONG nSize;

	nSize = (LONG)SendDlgItemMessage(hWnd, IDC_EDIT_LOG, WM_GETTEXTLENGTH, 0, 0);
	SendDlgItemMessage(hWnd, IDC_EDIT_LOG, EM_SETSEL, (WPARAM)nSize, (LPARAM)nSize);
	SendDlgItemMessage(hWnd, IDC_EDIT_LOG, EM_REPLACESEL, TRUE, (LPARAM)message);
	if (bNextLine == TRUE) {
		SendDlgItemMessage(hWnd, IDC_EDIT_LOG, EM_REPLACESEL, TRUE, (LPARAM)TEXT("\r\n"));
	}
}